;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Ave Cruce Salus Mea
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; program: Win3ds - demonstration of several graphics
;;          effect without openGL using.
;; author:  Maciej Guba aka macgub
;; web   :  www.macgub.hekko.pl
;; email :  macgub3@wp.pl
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Key description : SPACE -  switching
;;          beetwen displaying models: flat, smooth
;;          shading, spherical environment, texture,
;;          bump, smooth + texture, two texture and bump
;;          + texture mapping pararell; TAB - generate
;;          three unlinear lights (Phong illumination);
;;          +/- keys - zoom in/out; ESC - exit.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Progaram needs SSE2 extension.
format PE GUI 4.0
entry start

include '/include/win32a.inc'

XRES equ 800 ;1600 ;1200 ;800      ;choose screen size
YRES equ 450 ;900 ; 675 ;450
SIZE_X	equ XRES
SIZE_Y	equ YRES
TEX_X = 512    ; texture size
TEX_Y = 512
TEXTURE_SIZE EQU (TEX_X * TEX_Y)-1
TEX_SHIFT = 9
TRIANGLES_COUNT equ 4
POINTS_COUNT equ 4
ROUND equ 8
NON  = 0
MMX  = 1
SSE  = 2
SSE2 = 3
Ext  = SSE2


section '.text' code readable executable
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; start and message loop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
start:
    invoke GetModuleHandle,0
    mov [wc.hInstance],eax
    invoke LoadIcon,0,IDI_APPLICATION
    mov [wc.hIcon],eax
    invoke LoadCursor,0,IDC_ARROW
    mov [wc.hCursor],eax

    mov [bmi.biSize],sizeof.BITMAPINFOHEADER
    mov [bmi.biWidth],XRES
    mov [bmi.biHeight],-YRES
    mov [bmi.biPlanes],1
    mov [bmi.biBitCount],32
    mov [bmi.biCompression],BI_RGB

    mov [bmi.biSizeImage],0
    mov [bmi.biXPelsPerMeter],0
    mov [bmi.biYPelsPerMeter],0
    mov [bmi.biClrUsed],0
    mov [bmi.biClrImportant],0

    mov [rect.left],0
    mov [rect.top],0
    mov [rect.right],XRES
    mov [rect.bottom],YRES
    invoke AdjustWindowRect,rect,WS_POPUP or WS_SYSMENU or WS_CAPTION,0 

    invoke RegisterClass,wc

    mov ecx,[rect.bottom]
    mov eax,[rect.right]
    sub ecx,[rect.top]
    sub eax,[rect.left] 

    invoke CreateWindowEx,0,_class,_title,WS_OVERLAPPEDWINDOW\
	   xor WS_THICKFRAME xor WS_MAXIMIZEBOX,0,0,eax,ecx,NULL,NULL,0,0
    mov [hwnd],eax
    invoke GetDC,[hwnd]
    mov [hdc],eax
    invoke ShowWindow,[hwnd],SW_SHOW

    invoke timeGetTime
    mov [s_time],eax	     ;initialize start time
    pushad
    call   read_from_file

    call   init_triangles_normals2
    call   init_point_normals
    call   normalize_all_light_vectors
    call   init_envmap2
    call   do_color_buffer
    mov    edi,bump_map
    call   calc_bumpmap
    call   calc_bumpmap_coords
    call   generate_texture2
    popad
msg_loop:
    invoke timeGetTime
    sub eax,[s_time]
    mov [e_time],eax	     ;get new end time

    stdcall g_draw
    inc  [ccounter]
    and  [ccounter],0x000000ff
    invoke StretchDIBits,[hdc],0,0,XRES,YRES,0,0,XRES,YRES,buffer,bmi,0,SRCCOPY
  ;  invoke StretchDIBits,[hdc],0,0,TEX_X,TEX_Y,0,0,TEX_X,TEX_Y,color_map,bmi,0,SRCCOPY

    invoke Sleep,1

    invoke PeekMessage,msg,0,0,0,PM_NOREMOVE
    or eax,eax
    je msg_loop

    invoke GetMessage,msg,NULL,0,0
    or eax,eax
    je end_loop
    invoke TranslateMessage,msg
    invoke DispatchMessage,msg

    jmp msg_loop

error:
    invoke MessageBox,NULL,_error,NULL,MB_ICONERROR+MB_OK

end_loop:
    invoke ReleaseDC,[hwnd],[hdc]
    invoke DestroyWindow,[hwnd]
    invoke ExitProcess,[msg.wParam]
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; WindowProc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc WindowProc uses ebx esi edi, hwnd,wmsg,wparam,lparam
    cmp [wmsg],WM_KEYDOWN
    je .wmkeydown
    cmp [wmsg],WM_DESTROY
    je .wmdestroy
  .defwndproc:
    invoke DefWindowProc,[hwnd],[wmsg],[wparam],[lparam]
    jmp .finish
  .wmkeydown:
    cmp [wparam],VK_UP
    jne .vescape_or_set_shading

;    stdcall g_init
    xor eax,eax
    jmp .finish
  .vescape_or_set_shading:
    cmp [wparam],VK_SPACE
    jne @f				    ; possible shading models
    inc [draw_flag]			    ; - gouraud (smooth)
    cmp [draw_flag],8			    ; - flat
    jne  @f				    ; - flat optimized on sse2 with float Zbuffer
    mov [draw_flag],0
  @@:
    cmp   [wparam],VK_TAB
    jne   @f
    call  make_random_lights
    call  normalize_all_light_vectors
    call  do_color_buffer
    call  init_envmap2
;  @@:
;    cmp   [wparam],VK_F1
;    jne   @f
;    not   [blur_flag]
;    cmp   [blur_flag],0
;    je    @f
;    mov   ecx,2
;    call  blur_screen
  @@:
    cmp   [wparam],VK_ADD
    jne    @f
    fninit
    fld   [scale]
    fmul  [scale_plus]
    fstp  [scale]
  @@:
    cmp   [wparam],VK_SUBTRACT
    jne    @f
    fninit
    fld   [scale]
    fmul  [scale_minus]
    fstp  [scale]
  @@:
    cmp [wparam],VK_ESCAPE
    jne .finish
  .wmdestroy:
    invoke PostQuitMessage,0
    xor eax,eax
  .finish: 
    ret
endp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; g_draw - draw to screen buffer; main program proc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc g_draw ;uses ebx esi edi
local .temp dw ?, .rot_flag dw ?
;local .first_p_index dd ?, .second_p_index dd ?, .third_p_index dd ?
;local .zz1 dw ?, .zz2 dw ?, .zz3 dw ? ;-> don't change order
local .zz1f dd ?, .zz2f dd ?, .zz3f dd ?
local .xx1 dw ?, .xx2 dw ?, .xx3 dw ?
local .yy1 dw ?, .yy2 dw ?, .yy3 dw ?
local .xy1 dd ?, .xy2 dd ?, .xy3 dd ?
local .dpr dd ?, .dpr_p dd ?
local .x_temp_var dd ?, .y_temp_var dd ?
local .counter dd ?
  ;  invoke start_timing

    pushad
    cmp   [blur_flag],0
    jne   @f
    xor   eax,eax	   ; clear draw buffer
    mov   edi,buffer
    cld
    mov   ecx,XRES * YRES
    rep   stosd
  @@:

;    cmp   [draw_flag],2
;    je    @f
    mov   eax,5000*65536    ; make background to Zbuffer
;    jmp   .m_back
;  @@:
;    mov   eax,-6000.1
  .m_back:
    mov   edi,Zbuffer
    mov   ecx,XRES * YRES
    rep   stosd


  .rotate:
    fninit
    fild   [ccounter]
    fmul   [one_deg]
    fsincos
    fstp   [cos_beta]
    fstp   [sin_beta]
    mov    [.rot_flag],0
 .rot_points_and_normals:
    mov    esi,points_r 	; rotate points
    mov    edi,points_rotated
    jmp    .do_rot
 .rot_normals:
    mov    esi,points_normals
    mov    edi,points_normals_rotated
 .do_rot:
    movzx  ecx,[points_count_var]
 .again_rot:
    fld    dword[esi+4]
    fmul   [cos_beta]
    fld    dword[esi+8]
    fmul   [sin_beta]
    faddp
    fld    dword[esi+4]
    fmul   [sin_beta]
    fchs
    fld     dword[esi+8]
    fmul   [cos_beta]
    faddp
    fstp   dword[edi+8]
    fstp   dword[edi+4]
    push   dword[esi]
    pop    dword[edi]

    fld    dword[edi]
    fld    st
    fmul   [cos_beta]
    fld    dword[edi+8]
    fmul   [sin_beta]
    fchs
    faddp
    fstp   dword[edi]
    fmul   [sin_beta]
    fld    dword[edi+8]
    fmul   [cos_beta]
    faddp
    fstp   dword[edi+8]

    add    esi,12
    add    edi,12
  ;  cmp    dword[esi],-1
  ;  jnz    @b
    loop   .again_rot
    inc    [.rot_flag]
    cmp    [.rot_flag],1
    jle    .rot_normals

    push   dword[esi]
    pop    dword[edi]


    fninit

    mov   esi,triangles 	   ; draw triangles
    movzx ecx,[triangles_count_var]
  .draw_triangle:
    push  ecx
    fninit
    movzx edx,word[esi]
    shl   edx,2
    lea   edx,[edx*3]
    mov   [first_p_index],edx
    add   edx,points_rotated
    fld   dword[edx]
    fmul  [scale]
    fiadd [xxadd]
    fistp [.xx1]
    movzx eax,[.xx1]
    shl   eax,16
    fld   dword[edx+4]
    fmul  [scale]
    fiadd [yyadd]
    fistp [.yy1]
    mov   ax,[.yy1]
    fld   dword[edx+8]
    fmul  [scale]
    fiadd [zzadd]
    fist  word[zz1]
    fstp  [.zz1f]
					; z coord - skip
    movzx edx,word[esi+2]
    shl   edx,2
    lea   edx,[edx*3]
    mov   [second_p_index],edx
    add   edx,points_rotated
    fld   dword[edx]
    fmul  [scale]
    fiadd [xxadd]
    fistp [.xx2]
    movzx ebx,[.xx2]
    shl   ebx,16
    fld   dword[edx+4]
    fmul  [scale]
    fiadd [yyadd]
    fistp [.yy2]
    mov   bx,[.yy2]
    fld   dword[edx+8]
    fmul  [scale]
    fiadd [zzadd]
    fist  [zz2]
    fstp  [.zz2f]

    movzx edx,word[esi+4]
    shl   edx,2
    lea   edx,[edx*3]
    mov   [third_p_index],edx
    add   edx,points_rotated
    fld   dword[edx]
    fmul  [scale]
    fiadd [xxadd]
    fistp [.xx3]
    movzx ecx,[.xx3]
    shl   ecx,16
    fld   dword[edx+4]
    fmul  [scale]
    fiadd [yyadd]
    fistp [.yy3]
    mov   cx,[.yy3]
    fld   dword[edx+8]
    fmul  [scale]
    fiadd [zzadd]
    fist  [zz3]
    fstp  [.zz3f]

    mov     [.xy1],eax
    mov     [.xy2],ebx
    mov     [.xy3],ecx

    push  esi
    pushad

    cmp   [draw_flag],0
    je	  .flat_draw
    cmp   [draw_flag],2
    je	  .env_mapping
;    je    .flat_draw_float
;    cmp   [draw_flag],3
;    je    .env_mapping
    cmp   [draw_flag],3
    je	  .bump_mapping
    cmp   [draw_flag],4
    je	  .tex_mapping
    cmp   [draw_flag],7
    je	  .bump_tex_mapping
    cmp   [draw_flag],6
    je	  .two_tex
    cmp   [draw_flag],5
    je	  .grd_tex


     .grd_draw: 	       ;****Gouraud shading****
	fninit


	mov	esi,third_p_index
	mov	ecx,3
      .again_grd_draw:
	mov	eax,[esi]
	add	eax,points_normals_rotated
	fld	dword[eax]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	[.x_temp_var]	 ; x temp variables
	fld	dword[eax+4]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	[.y_temp_var]

	mov	eax,[.y_temp_var]
	shl	eax,TEX_SHIFT
	add	eax,[.x_temp_var]
	lea	eax,[4*eax+color_map]
	mov	eax,dword[eax]
	push	word[ecx*2+zz1-2]


	ror	 eax,16 	      ; eax -0xxxrrggbb -> 0xggbbxxrr
	xor	 ah,ah
	push	 ax	    ;r
	rol	 eax,8		      ; eax-0xggbb00rr -> 0xbb00rrgg
	xor	 ah,ah
	push	 ax	    ;g
	shr	 eax,24
	push	 ax	    ;b

	sub	 esi,4
	dec	 ecx
	jnz	 .again_grd_draw

if 0
   ; -----------version with only one light supported------
    mov   [.counter],4	    ; ++++ gouraud shading  ++++++++
  .calc_col_LVD:    ; calculate color Light Vector Depend
    mov   edx,[.counter]
    push  [zz1+edx]
; calc color of triangle light vector dependent
    mov   esi,light_vect
    mov   edi,points_normals_rotated
    add   edi,[first_p_index+2*edx]
    call  dot_product
    fst   [.dpr]	    ; orginal+specular*st+emmisive*st^5
    push  ecx		    ; o + min*st + max * st^5
    mov   ecx,5 	    ; shiness
  @@:
    fmul  st,st0
    loop  @b
    fstp  [.dpr_p]	    ; dot product powered
    pop   ecx


    movq      xmm0,[light_c_max]
    pxor      xmm1,xmm1
    punpcklwd xmm0,xmm1
    cvtdq2ps  xmm2,xmm0
    movss     xmm3,[.dpr_p]
    shufps    xmm3,xmm3,0
    mulps     xmm3,xmm2

    movq      xmm0,[light_c_o]
    punpcklwd xmm0,xmm1
    cvtdq2ps  xmm2,xmm0
    addps     xmm3,xmm2

    movq      xmm0,[light_c_min]
    punpcklwd xmm0,xmm1
    cvtdq2ps  xmm2,xmm0
    movss     xmm4,[.dpr]
    shufps    xmm4,xmm4,0
    mulps     xmm2,xmm4
    addps     xmm3,xmm2

;    shufps    xmm3,xmm3,11000110b

    cvtps2dq  xmm3,xmm3
    packssdw  xmm3,xmm3
;    pmaxuw    xmm3,[minimum0]
;    pminuw    xmm3,[maximum255]

    movd      edx,xmm3
  ;  ror       edx,16
    push      edx
    psrldq    xmm3,4
    movd      edx,xmm3
    push      dx

    sub   [.counter],2
    cmp   [.counter],0
    jge   .calc_col_LVD


end if

    mov     eax,[.xy1]
    mov     ebx,[.xy2]
    mov     ecx,[.xy3]

    mov   esi,Zbuffer
    mov   edi,buffer
    call  gouraud_triangle_z
    jmp   .end_draw


    .grd_tex:


	fninit

	mov    eax,[third_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[second_p_index]	 ; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[first_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov	esi,third_p_index
	mov	ecx,3

      .again_grd_tex_draw:

	mov	eax,[esi]
	add	eax,points_normals_rotated
	fld	dword[eax]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	[.x_temp_var]	 ; x temp variables
	fld	dword[eax+4]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	[.y_temp_var]

	mov	eax,[.y_temp_var]
	shl	eax,TEX_SHIFT
	add	eax,[.x_temp_var]
	lea	eax,[4*eax+color_map]
	mov	eax,dword[eax]
	push	word[ecx*2+zz1-2]


	ror	 eax,16 	      ; eax -0xxxrrggbb -> 0xggbbxxrr
	xor	 ah,ah
	push	 ax	    ;r
	rol	 eax,8		      ; eax-0xggbb00rr -> 0xbb00rrgg
	xor	 ah,ah
	push	 ax	    ;g
	shr	 eax,24
	push	 ax	    ;b

	sub	 esi,4
	dec	 ecx
	jnz	 .again_grd_tex_draw


	mov    eax,[.xy1]
	mov    ebx,[.xy2]
	mov    ecx,[.xy3]

	mov	edi,buffer   ; screen buffer
	mov	esi,Zbuffer
	mov	edx,texmap

	call	tex_plus_grd_triangle

	jmp    .end_draw



     .flat_draw:		     ;**************************                                     ; FLAT DRAWING
	fninit
	push	eax
	push	ebx
	push	ecx

	mov	eax,[first_p_index]
	add	eax,points_normals_rotated
	fld	dword[eax]
	mov	ebx,[second_p_index]
	add	ebx,points_normals_rotated
	fadd	dword[ebx]
	mov	ecx,[third_p_index]
	add	ecx,points_normals_rotated
	fadd	dword[ecx]

	fidiv	[i3]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	dword[.x_temp_var] ;[esp-4]    ; x temp variables
	fld	dword[eax+4]	; y cooficient of normal vector
	fadd	dword[ebx+4]
	fadd	dword[ecx+4]
	fidiv	[i3]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	dword[.y_temp_var];[esp-8]   ;  y
	mov	edx,[.x_temp_var]  ; dword[esp-8]
	shl	edx,TEX_SHIFT
	add	edx,[.y_temp_var] ;dword[esp-4]
	lea	eax,[4*edx+color_map]
	mov	edx,dword[eax]
	or	edx,0xff000000	 ; edx - color

	pop	ecx
	pop	ebx
	pop	eax

	lea	edi,[buffer]
	mov	esi,Zbuffer
	push	word[zz3]
	push	word[zz2]
	push	word[zz1]

	call	flat_triangle_z

    jmp   .end_draw






    .env_mapping:
	push	[zz3]
	push	[zz2]
	push	word[zz1]
	fninit
	xor	esi,esi
	sub	esp,12
	mov	edi,esp
	mov	[.counter],3
      @@:
	mov	edx,[esi+first_p_index]
	add	edx,points_normals_rotated    ;point_normals_rotated
	; texture x=(rotated point normal -> x * 255)+255
	fld	dword[edx]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi]
	; texture y=(rotated point normal -> y * 255)+255
	fld	dword[edx+4]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi+2]

	add	edi,4
	add	esi,4
	dec	[.counter]
	jnz	@b

	mov	edi,buffer   ; screen buffer
	mov	esi,envmap
	mov	edx,Zbuffer
  ;      mov     edx,[Zbuffer_ptr]
	call	tex_triangle_z
	jmp	.end_draw

if 0
 .flat_draw_float:
    push  [.zz3f]
    push  [.zz2f]
    push  [.zz1f]

; calc color of triangle light vector dependent

    mov   esi,light_vect
    mov   edi,points_normals_rotated
    add   edi,[first_p_index]
    call  dot_product
    fst   [.dpr]	    ; orginal+specular*st+emmisive*st^5
    push  ecx		    ; o + min*st + max * st^5
    mov   ecx,5 	    ; shiness
  @@:
    fmul  st,st0
    loop  @b
    fstp  [.dpr_p]	    ; dot product powered
    pop   ecx

    movq      xmm0,[light_c_max]
    pxor      xmm1,xmm1
    punpcklwd xmm0,xmm1
    cvtdq2ps  xmm2,xmm0
    movss     xmm3,[.dpr_p]
    shufps    xmm3,xmm3,0
    mulps     xmm3,xmm2

    movq      xmm0,[light_c_o]
    punpcklwd xmm0,xmm1
    cvtdq2ps  xmm2,xmm0
    addps     xmm3,xmm2

    movq      xmm0,[light_c_min]
    punpcklwd xmm0,xmm1
    cvtdq2ps  xmm2,xmm0
    movss     xmm4,[.dpr]
    shufps    xmm4,xmm4,0
    mulps     xmm2,xmm4
    addps     xmm3,xmm2

    cvtps2dq  xmm3,xmm3
    packssdw  xmm3,xmm3
;    pmaxub    xmm3,[minimum0]
;    pminub    xmm3,[maximum255]
    packuswb  xmm3,xmm3
    movd      edx,xmm3

    mov   esi,Zbuffer
    mov   edi,buffer
;    movzx edx,[.zz1]

;    sub   edx,90
;    mov   dh,dl
  ;  mov   dh,dl
  ;  movzx edx,dx
    call  flat_triangle_z_float
    jmp     .end_draw
end if

  .tex_mapping:
	push	[zz3]
	push	[zz2]
	push	word[zz1]

	mov    eax,[third_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[second_p_index]	 ; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[first_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[.xy1]
	mov    ebx,[.xy2]
	mov    ecx,[.xy3]

	mov	edi,buffer   ; screen buffer
	mov	esi,texmap
	mov	edx,Zbuffer
	call	tex_triangle_z



	jmp	.end_draw


    .bump_mapping:
	push	Zbuffer
	push	word[zz3]
	push	word[zz2]
	push	word[zz1]

	fninit
	xor	esi,esi
	sub	esp,12
	mov	edi,esp
	mov	[.counter],3
      @@:
	mov	edx,[esi+first_p_index]
	add	edx,points_normals_rotated    ;point_normals_rotated
	; texture x=(rotated point normal -> x * 255)+255
	fld	dword[edx]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi]
	; texture y=(rotated point normal -> y * 255)+255
	fld	dword[edx+4]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi+2]

	add	edi,4
	add	esi,4
	dec	[.counter]
	jnz	@b

	mov    eax,[third_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[second_p_index]	 ; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[first_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[.xy1]
	mov    ebx,[.xy2]
	mov    ecx,[.xy3]

	mov	edi,buffer
	mov	esi,envmap
	mov	edx,bump_map		;BUMP_MAPPING

	call	bump_triangle_z
	jmp    .end_draw

    .two_tex:		    ; double texture mapping

	push	Zbuffer
	push	word[zz3]
	push	word[zz2]
	push	word[zz1]

	fninit
	xor	esi,esi
	sub	esp,12
	mov	edi,esp
	mov	[.counter],3
      @@:
	mov	edx,[esi+first_p_index]
	add	edx,points_normals_rotated    ;point_normals_rotated
	; texture x=(rotated point normal -> x * 255)+255
	fld	dword[edx]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi]
	; texture y=(rotated point normal -> y * 255)+255
	fld	dword[edx+4]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi+2]

	add	edi,4
	add	esi,4
	dec	[.counter]
	jnz	@b

	mov    eax,[third_p_index]	; tex coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[second_p_index]	 ; tex coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[first_p_index]	; tex coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[.xy1]
	mov    ebx,[.xy2]
	mov    ecx,[.xy3]

	mov	edi,buffer
	mov	esi,envmap
	mov	edx,texmap

	call	two_tex_triangle_z
	jmp    .end_draw


    .bump_tex_mapping:

	mov    eax,[third_p_index]	; tex coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[second_p_index]	 ; tex coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[first_p_index]	; tex coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	push	texmap

	push	Zbuffer
	push	word[zz3]
	push	word[zz2]
	push	word[zz1]

	fninit
	xor	esi,esi
	sub	esp,12
	mov	edi,esp
	mov	[.counter],3
      @@:
	mov	edx,[esi+first_p_index]
	add	edx,points_normals_rotated    ;point_normals_rotated
	; texture x=(rotated point normal -> x * 255)+255
	fld	dword[edx]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi]
	; texture y=(rotated point normal -> y * 255)+255
	fld	dword[edx+4]
	fimul	[correct_tex]
	fiadd	[correct_tex]
	fistp	word[edi+2]

	add	edi,4
	add	esi,4
	dec	[.counter]
	jnz	@b

	mov    eax,[third_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[second_p_index]	 ; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[first_p_index]	; bump map coords
	cdq
	idiv   [i3]
	add    eax,tex_points
	push   dword[eax]

	mov    eax,[.xy1]
	mov    ebx,[.xy2]
	mov    ecx,[.xy3]

	mov	edi,buffer
	mov	esi,envmap
	mov	edx,bump_map		;BUMP_MAPPING

	call	bump_tex_triangle_z






 .end_draw:
    popad
    pop   esi
    add   esi,6
;    cmp   dword[esi],-1
;    jnz   .draw_triangle
    pop    ecx
    dec    ecx
    jnz    .draw_triangle

    popad
;    rdtsc
;    pop    ebx
;    sub    eax,ebx

;    mov     ecx,10
;  .dc:
;    xor     edx,edx
;    mov     edi,10
;    div     edi
;    add     dl,30h
;    mov     [STRdata+ecx-1],dl
;    loop    .dc

   ; invoke      MessageBox,[hwnd],STRdata,_tsc_title,MB_RTLREADING
;   invoke stop_ticks
ret

endp

normalize_all_light_vectors:
	mov	edi,lights
     @@:
	call	normalize_vector	   ;       3dmath.inc
	add	edi,LIGHT_SIZE
	cmp	edi,lightsend	;ecx
	jl	@b
ret

read_from_file:
	fninit
    ;    mov     edi,[triangles_ptr]
	mov	edi,triangles
	xor	ebx,ebx
	xor	ebp,ebp
	mov	[points_count_var],0
	mov	[triangles_count_var],0
  ; if USE_LFN = 0
	mov	esi,SourceFile
  ; else
  ;      mov     esi,[fptr]
  ; end if
	cmp	[esi],word 4D4Dh
	jne	.exit ;Must be legal .3DS file
;        cmp     dword[esi+2],EndFile-SourceFile
;        jne     .exit ;This must tell the length
	mov	eax,dword[esi+2]
  ;      cmp     eax,[fsize]
  ;      jne     .exit

	add	eax,esi
	mov	[EndFile],eax	 ;

	add	esi,6
      @@:
	cmp	[esi],word 3D3Dh
	je	@f
	add	esi,[esi+2]
	jmp	@b
      @@:
	add	esi,6
      .find4k:
	cmp	[esi],word 4000h
	je	@f
	add	esi,[esi+2]
	cmp	esi,[EndFile]
	jc	.find4k
	jmp	.exit
      @@:
	add	esi,6
      @@:
	cmp	[esi],byte 0
	je	@f
	inc	esi
	jmp	@b
      @@:
	inc	esi
      @@:
	cmp	[esi],word 4100h
	je	@f
	add	esi,[esi+2]
	jmp	@b
      @@:
	add	esi,6
      @@:
	cmp	[esi],word 4110h
	je	@f
	add	esi,[esi+2]
	jmp	@b
      @@:
	movzx	ecx,word[esi+6]
	add	[points_count_var],cx

	mov	edx,ecx
	add	esi,8
     @@:
	push	edi
    ;    mov     edi,[points_ptr]
	mov	edi,points_r
	push	dword[esi+4]
	pop	dword[edi+ebx*2+0]
	push	dword[esi+8]
	pop	dword[edi+ebx*2+4]
	push	dword[esi+0]
	pop	dword[edi+ebx*2+8]
	pop	edi
;        fld     dword[esi+4]
;        fstp    dword[real_points+ebx*2+0]  ; x
;        fld     dword[esi+8]
;        fstp   dword[real_points+ebx*2+4]  ; y
;        fld     dword[esi+0]
;        fstp   dword[real_points+ebx*2+8]  ; z

	add	ebx,6
	add	esi,12
	dec	ecx
	jnz	@b
      @@:
  ;      mov     dword[points+ebx],-1
	push	edi
    ;    mov     edi,[points_ptr]
	mov	edi,points_r
	mov	dword[edi+ebx*2],-1	   ; end mark (not always in use)
	pop	edi
      @@:
	cmp	[esi],word 4120h
	je	@f
	add	esi,[esi+2]
	jmp	@b
      @@:
	movzx	ecx,word[esi+6]
	add	[triangles_count_var],cx
	add	esi,8
	;mov     edi,triangles
      @@:
	movsd
	movsw
	add	word[edi-6],bp
	add	word[edi-4],bp
	add	word[edi-2],bp
	add	esi,2
	dec	ecx
	jnz	@b
	add	ebp,edx
	jmp	.find4k
	mov	eax,-1 ;<---mark if OK
      .exit:
	mov	dword[edi],-1
ret

init_triangles_normals2:
    ;    mov     ebx,[triangles_normals_ptr]
	mov	ebx,triangles_normals
    ;    mov     ebp,[triangles_ptr]
	mov	ebp,triangles
	movzx	ecx,[points_count_var]
     @@:
	push	ecx
	push	ebx
	mov	ebx,vectors
	movzx	esi,word[ebp]	       ; first point index
	lea	esi,[esi*3]
    ;    lea     esi,[points+esi*2]     ; esi - pointer to 1st 3d point
	shl	esi,2
	add	esi,points_r
	movzx	edi,word[ebp+2] 	 ; first point index
	lea	edi,[edi*3]
	shl	edi,2
  ;      add     edi,[points_ptr]
	add	edi,points_r
;        movzx   edi,word[ebp+2]        ; second point index
;        lea     edi,[edi*3]
;        lea     edi,[points+edi*2]     ; edi - pointer to 2nd 3d point

	call	make_vector_r
	add	ebx,12
	mov	esi,edi
	movzx	edi,word[ebp+4]        ; third point index
	lea	edi,[edi*3]
	shl	edi,2
      ;  add     edi,[points_ptr]
	add	edi,points_r
;        lea     edi,[points+edi*2]
	call	make_vector_r
	mov	edi,ebx 		; edi - pointer to 2nd vector
	mov	esi,ebx
	sub	esi,12			; esi - pointer to 1st vector
	pop	ebx
	call	cross_product
	mov	edi,ebx
	call	normalize_vector
	add	ebp,6
	add	ebx,12
	pop	ecx
    ;    loop    @b    why loop version don't works correct ?
	cmp	dword[ebp],-1
	jne	@b
ret

init_point_normals:
.x equ dword [ebp-4]
.y equ dword [ebp-8]
.z equ dword [ebp-12]
.point_number equ word [ebp-26]
.hit_faces    equ word [ebp-28]

	fninit
	mov	  ebp,esp
	sub	  esp,28
	mov	  edi,points_normals ;[points_normals_ptr]
	mov	  .point_number,0
    .ipn_loop:
	mov	  .hit_faces,0
	mov	  .x,0
	mov	  .y,0
	mov	  .z,0
	mov	  esi,triangles;[triangles_ptr]
	xor	  ecx,ecx	       ; ecx - triangle number
    .ipn_check_face:
	xor	  ebx,ebx	       ; ebx - 'position' in one triangle
    .ipn_check_vertex:
	movzx	  eax,word[esi+ebx]    ;  eax - point_number
	cmp	  ax,.point_number
	jne	  .ipn_next_vertex
	push	  esi
	mov	  esi,ecx
	lea	  esi,[esi*3]
       ; lea       esi,[triangles_normals+esi*4]
	shl	  esi,2
	add	  esi,triangles_normals ;[triangles_normals_ptr]

	fld	  .x
	fadd	  dword[esi]
	fstp	  .x
	fld	  .y
	fadd	  dword[esi+4]
	fstp	  .y
	fld	  .z
	fadd	  dword[esi+8]
	fstp	  .z
	pop	  esi
	inc	  .hit_faces
	jmp	  .ipn_next_face
    .ipn_next_vertex:
	add	  ebx,2
	cmp	  ebx,6
	jne	  .ipn_check_vertex
    .ipn_next_face:
	add	  esi,6
	inc	  ecx
	cmp	  cx,[triangles_count_var]
	jne	  .ipn_check_face

	fld	  .x
	fidiv	  .hit_faces
	fstp	  dword[edi]
	fld	  .y
	fidiv	  .hit_faces
	fstp	  dword[edi+4]
	fld	  .z
	fidiv	  .hit_faces
	fstp	  dword[edi+8]
	call	  normalize_vector
	add	  edi,12  ;type vector 3d
	inc	  .point_number
	mov	  dx,.point_number
	cmp	  dx,[points_count_var]
	jne	  .ipn_loop

	mov	  esp,ebp
ret

include 'grd_cat.inc' ;'grd_cat_opt.inc'
include 'tex_cat.inc'
include 'flat_cat_opt3.inc'
;include 'flat_cat_opt3_floatZ.inc'
include '3DMATH.INC'
include 'procs_a.inc'
include 'bump_cat.inc'
include 'bump_tex.inc'
include 'two_tex.inc'
include 'grd_tex.inc'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; data
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
section '.data' data readable writeable
    _class TCHAR     'Win3ds',0
    _title TCHAR     'Ave Cruce Salus Mea.',0
    _error TCHAR     'Startup failed.',0
    _tsc_title TCHAR 'Clocks number',0
    SourceFile:
    file	'knot.3ds'
    EndFile	dd ?
    up		dd 0
    s_time	dd 0
    e_time	dd 0
    num_animations dd 3
    ccounter	   dd 0
    triangles_count_var  dw TRIANGLES_COUNT
    points_count_var	 dw POINTS_COUNT

 ;   points_r:
 ;     dd -1.0, -1.0, 1.0
 ;     dd  1.0, -1.0, 1.0
 ;     dd  0.0,  1.0, 0.0
 ;     dd  0.0, -1.0, -1.0
 ;     dd -1 ;<- end marker

 ;   triangles:
 ;      dw 0,2,1, 3,2,0, 1,2,3,  0,1,3
 ;      dd -1 ;<- end marker

    tex_x_div2	dd  TEX_X / 2
    tex_y_div2	dd  TEX_Y / 2
    i3		dd  3
    i256	dw  256
    i255d	dd  255
    env_const	dd  1.05
    dot_max	dd  1.0
    dot_min	dd  0.0
    correct_tex dw  255


    scale	dd 10.0
    scale_plus	dd 1.2
    scale_minus dd 0.8
    xxadd      dw XRES/2
    yyadd      dw YRES/2
    zzadd      dw 1000
    draw_flag  db 0
    blur_flag  db 0
    piD128:
    one_deg	dd   0.024543692
    STRdata	db   '-1        '
    all_lights_size dw lightsend-lights

  lights:
    light_vect	dd   0.0,0.0,-1.0
    light_c_o:			   ;   +12
       dw   1,1,1,0
    light_c_min:		   ;   +20
       dw   1,1,1,0
    light_c_max:		   ;   +28
       dw   255,255,5,0
    shiness:
       dw   5			   ;   +36
    LIGHT_SIZE = ($-lights)
       dd   -0.5,-0.5,-1.0
			 ;   +12
       dw   1,255,1,0
			 ;   +20
       dw   1,1,1,0
			 ;   +28
       dw   255,255,255,0
       dw   5		 ;   +36

       dd   0.5,0.5,-1.0
				   ;   +12
       dw   13,1,1,0
				   ;   +20
       dw   70,30,70,0
				   ;   +28
       dw   25,255,255,0
       dw   5			   ;   +36
   lightsend:


 align 16
    minimum0:
       dw   0,0,0,0,0,0,0,0
    maximum255:
       dw   255,255,255,255,255,255,255,255
 align 16
    mask_blend:
       dd -1,-1,-1,-1
    blur_sub:
       dd 0x01010101, 0x01010101, 0x01010101, 0x01010101
 align 16
    muler1234:
       dd 1.0
       dd 2.0
       dd 3.0
       dd 4.0
align 16
    muler0123:
       dd 0.0
       dd 1.0
       dd 2.0
       dd 3.0
    msg  MSG
    rect RECT

    wc WNDCLASS 0,WindowProc,0,0,NULL,NULL,NULL,COLOR_BTNFACE+1,NULL,_class
    cos_beta  dd ?
    sin_beta  dd ?
    rand_seed dw ?
    hwnd      dd ?
    hdc       dd ?
    bmi      BITMAPINFOHEADER
    align 16
    sse_repository:
       rd 512
  align 16
    buffer    rd XRES * YRES
    Zbuffer   rd XRES * YRES
    bump_map  rb TEX_X * TEX_Y
    envmap    rd TEX_X * TEX_Y
    color_map rd TEX_X * TEX_Y
    texmap    rd TEX_X * TEX_Y
    points_rotated rb 100000
    tex_points	rb 100000	; bump_map  and texture coords
				; each point word x, word y
    points_r:
       rb 100000
    triangles:
       rb 100000
    triangles_normals:
       rb 100000
    points_normals:
       rb 100000
    points_normals_rotated:
       rb 100000

    first_p_index:
		   dd ?
    second_p_index dd ?
    third_p_index  dd ?   ; don't change order
    zz1:
	dw ?
    zz2 dw ?
    zz3 dw ? ;-> don't change order

    vectors rb 24
    matrix  rb 36
    cos_tab rd 360
    sin_tab rd 360


section '.idata' import data readable writeable
    library kernel32,'KERNEL32',\user32,'USER32',\gdi32,'GDI32',\winmm,'WINMM';,\
	 ;   timing,'timing.dll'
    include '\include\api\kernel32.inc'
    include '\include\api\user32.inc'
    include '\include\api\gdi32.inc'
 ;   include 'timing.inc'
    import winmm,timeGetTime,'timeGetTime'
